# jobs.py - Collection of workerpool Job objects
# Copyright (c) 2008 Andrey Petrov
#
# This module is part of s3funnel and is released under
# the MIT license: http://www.opensource.org/licenses/mit-license.php

from workerpool import Job
import boto
import time
try:
    import hashlib
except ImportError:
    import md5 as hashlib

import os
import logging
log = logging.getLogger(__name__)

READ_CHUNK = 8192

# Various exceptions we're expecting
from boto.exception import BotoServerError, BotoClientError, S3ResponseError
from httplib import IncompleteRead
from socket import error as SocketError
from s3funnel import FunnelError

class JobError(Exception):
    pass

# Jobs

class GetJob(Job):
    "Download the given key from S3."
    def __init__(self, bucket, key, failed, config={}):
        self.bucket = bucket
        self.key = key
        self.failed = failed
        self.retries = config.get('retry', 5)
        self.ignore_s3fs_dirs = config.get('ignore_s3fs_dirs',True)

    def _do(self, toolbox):
        for i in xrange(self.retries):
            try:
                b = toolbox.get_bucket(self.bucket)
                    
                b.connection.provider.metadata_prefix = ''
                k = b.get_key(self.key)
                m = k.get_metadata('content-type')
                
                if m == 'application/x-directory' and self.ignore_s3fs_dirs:
                     log.warn("Skipping s3fs directory: %s" % self.key)
                     return
                try:
                    # Create directories in case key has "/"
                    if os.path.dirname(self.key) and not os.path.exists(os.path.dirname(self.key)):
                        os.makedirs(os.path.dirname(self.key))
                except OSError:
                    pass
                # Note: This creates a file, even if the download fails
                k.get_contents_to_filename(self.key)
                log.info("Got: %s" % self.key)
                return
            except S3ResponseError, e:
                if e.status == 404:
                    log.error("Not found: %s" % self.key)
                    return
                else:
                    log.warning("Connection lost, reconnecting and retrying...")
                    toolbox.reset()
            except BotoServerError, e:
                break
            except (IncompleteRead, SocketError, BotoClientError), e:
                log.warning("Caught exception: %r.\nRetrying..." % e)
                time.sleep((2 ** i) / 4.0) # Exponential backoff
            except IOError, e:
                log.error("%s: '%s'" % (e.strerror, e.filename))
                return

        log.error("Failed to get: %s" % self.key)
        raise JobError()

    def run(self, toolbox):
        try:
            self._do(toolbox)
        except JobError, e:
            os.unlink(self.key) # Remove file since download failed
            self.failed.put(self.key)
        except Exception, e:
            self.failed.put(e)

class PutJob(Job):
    "Upload the given file to S3, where the key corresponds to basename(path)"
    def __init__(self, bucket, path, failed, config={}):
        self.bucket = bucket
        self.path = path
        self.failed = failed
        # --add-prefix logic
        self.add_prefix = config.get('add_prefix', '')
        self.key = "%s%s" % (self.add_prefix, self.path)
        # --del-prefix logic
        self.del_prefix = config.get('del_prefix')
        if self.del_prefix and self.key.startswith(self.del_prefix): 
            self.key = self.key.replace(self.del_prefix, '', 1)    
        if not config.get('put_full_path'):
            self.key = os.path.basename(self.key)
        self.retries = config.get('retry', 5)
        self.only_new = config.get('put_only_new')
        self.headers = config.get('headers', {})
        acl = config.get('acl')
        if acl not in ['private', 'public-read', 'public-read-write', 'authenticated-read']:
            log.warning("Bad ACL `%s` for key, setting to `private`: %s" % (self.acl, self.key))
            acl = 'private'
        self.headers['x-amz-acl'] = acl

    def _is_new(self, bucket, key):
        # Get existing key etag
        k = bucket.get_key(key)
        if not k: return True
        etag = k.etag[1:-1]

        # Compute file md5
        fp = open(self.path, 'rb')
        hash = hashlib.md5()
        data = fp.read(READ_CHUNK)
        while data:
            hash.update(data)
            data = fp.read(READ_CHUNK)
        fp.close()
        digest = hash.hexdigest()

        return etag != digest

    def _do(self, toolbox):
        for i in xrange(self.retries):
            try:
                bucket = toolbox.get_bucket(self.bucket)
                if self.only_new and not self._is_new(bucket, self.key):
                    log.info("Already exists, skipped: %s" % self.key)
                    return

                k = bucket.new_key(self.key)
                k.set_contents_from_filename(self.path, self.headers)
                log.info("Sent: %s" % self.key)
                return
            except S3ResponseError, e:
                log.warning("Connection lost, reconnecting and retrying...")
                toolbox.reset()
            except (IncompleteRead, SocketError, BotoClientError), e:
                log.warning("Caught exception: %r.\nRetrying..." % e)
                time.sleep((2 ** i) / 4.0) # Exponential backoff
            except IOError, e:
                log.warning("Path does not exist, skipping: %s" % self.path)
                break
            except Exception, e:
                log.critical("Unexpected exception: %r" % e)
                break

        log.error("Failed to put: %s" % self.key)

    def run(self, toolbox):
        try:
            self._do(toolbox)
        except JobError, e:
            self.failed.put(self.key)
        except Exception, e:
            self.failed.put(e)

class DeleteJob(Job):
    "Delete the given key from S3."
    def __init__(self, bucket, key, failed, config={}):
        self.bucket = bucket
        self.key = key
        self.failed = failed
        self.retries = config.get('retry', 5)

    def _do(self, toolbox):
        for i in xrange(self.retries):
            try:
                k = toolbox.get_bucket(self.bucket).delete_key(self.key)
                log.info("Deleted: %s" % self.key)
                return
            except S3ResponseError, e:
                log.warning("Connection lost, reconnecting and retrying...")
                toolbox.reset()
            except BotoServerError, e:
                break
            except (IncompleteRead, SocketError, BotoClientError), e:
                log.warning("Caught exception: %r.\nRetrying..." % e)
                time.sleep((2 ** i) / 4.0) # Exponential backoff

        log.error("Failed to delete: %s" % self.key)

    def run(self, toolbox):
        try:
            self._do(toolbox)
        except JobError, e:
            self.failed.put(self.key)
        except Exception, e:
            self.failed.put(e)
            
class CopyJob(Job):
    "Copy the given key from another bucket."
    def __init__(self, bucket, key, failed, config={}):
        self.bucket = bucket
        self.key = key
        # --add-prefix logic
        self.add_prefix = config.get('add_prefix', '')
        self.dest_key = "%s%s" % (self.add_prefix, key)
        # --del-prefix logic
        self.del_prefix = config.get('del_prefix')
        if self.del_prefix and self.dest_key.startswith(self.del_prefix): 
            self.dest_key = self.dest_key.replace(self.del_prefix, '', 1)
        self.source_bucket = config.get('source_bucket')
        self.failed = failed
        self.retries = config.get('retry', 5)
        
    def _do(self, toolbox):
        for i in xrange(self.retries):
            try:
                k = toolbox.get_bucket(self.bucket).copy_key(self.dest_key, self.source_bucket, self.key)
                log.info("Copied: %s to %s" % (self.key, self.dest_key))
                return
            except S3ResponseError, e:
                log.warning("Connection lost, reconnecting and retrying...")
                toolbox.reset()
            except BotoServerError, e:
                break
            except (IncompleteRead, SocketError, BotoClientError), e:
                log.warning("Caught exception: %r.\nRetrying..." % e)
                time.sleep((2 ** i) / 4.0) # Exponential backoff

        log.error("Failed to copy: %s" % self.key)
        
    def run(self, toolbox):
        try:
            self._do(toolbox)
        except JobError, e:
            self.failed.put(self.key)
        except Exception, e:
            self.failed.put(e)        



import logging
log = logging.getLogger(__name__)

import boto
import workerpool

from boto.exception import BotoServerError, BotoClientError, S3ResponseError
from httplib import IncompleteRead
from socket import error as SocketError
from Queue import Queue, Empty
from exceptions import FunnelError

from jobs import GetJob, PutJob, DeleteJob, CopyJob

__all__ = ['GetJob','PutJob','DeleteJob','S3ToolBox','BucketFunnel']

# Helpers

def collapse_queue(q):
    "Given a queue, return all the items currently in it."
    items = []
    try:
        while 1:
            i = q.get(block=False)
            q.task_done()
            if isinstance(i, FunnelError):
                raise i
            items.append(i)
    except Empty:
        return items

# Factories used to instantiate a workerpool

class S3ToolBox(object):
    """
    Container object for resources needed to access S3.
    This includes a connection to S3 and an instance of the bucket.
    """
    def __init__(self, aws_key, aws_secret_key, secure):
        self.aws_key = aws_key
        self.aws_secret_key = aws_secret_key
        self.secure = secure

        self.reset()

    def reset(self):
        self.conn = None
        self.buckets = {}

    def get_conn(self):
        "Get an S3 connection instance (cached)"
        if self.conn: return self.conn

        log.debug("Starting new S3 connection.")
        self.conn = boto.connect_s3(self.aws_key, self.aws_secret_key, is_secure=self.secure)
        return self.conn

    def get_bucket(self, name):
        "Get an S3 bucket instance with the given ``name`` (cached)"
        bucket = self.buckets.get(name)
        if bucket: return bucket

        conn = self.get_conn()
        log.debug("Getting bucket instance: %s" % name)
        try:
            bucket = conn.get_bucket(name)
        except BotoServerError, e:
            raise FunnelError("Bucket not found: %s" % name, key=name)

        self.buckets[name] = bucket
        return bucket

# Abstraction objects

class S3Funnel(object):
    """
    Abstraction object for manipulating S3.

    This class is not thread-safe but individual methods will use multiple
    threads to accomplish the request (where appropriate).

    If a ``pool`` is not given then a workerpool will be created at the first
    need of one. Optional pool config settings include:

    numthreads [Default: 5]
        Number of threads to create the pool with

    maxjobs [Default: numthreads * 2]
        Number of jobs to accept into the queue at a time (block if full)
    """
    def __init__(self, aws_key=None, aws_secret_key=None, pool=None, **config):
        self.aws_key = aws_key or config.get('aws_key')
        self.aws_secret_key = aws_secret_key or config.get('aws_secret_key')

        self.config = config
        self.numthreads = config.get('numthreads', 5)
        self.maxjobs = config.get('maxjobs', self.numthreads*2)
        self.secure = config.get('secure', True)

        self.pool = pool
        self.conn = None
        self.buckets = {}

        toolbox = S3ToolBox(self.aws_key, self.aws_secret_key, self.secure)
        self._get_conn = toolbox.get_conn
        self._get_bucket = toolbox.get_bucket

    def shutdown(self):
        if self.pool:
            self.pool.shutdown()

    def _get_pool(self):
        "Get a worker pool (cached)"
        if self.pool: return self.pool

        def toolbox_factory():
            return S3ToolBox(self.aws_key, self.aws_secret_key, self.secure)
        def worker_factory(job_queue):
            return workerpool.EquippedWorker(job_queue, toolbox_factory)

        log.info("Starting pool with %d threads." % self.numthreads)
        self.pool = workerpool.WorkerPool(self.numthreads, maxjobs=self.maxjobs, worker_factory=worker_factory)
        return self.pool

    def show_buckets(self):
        "Return an iterator of all the available bucket names"
        conn = self._get_conn()
        return (i.name for i in conn.get_all_buckets())

    def create_bucket(self, name, **config):
        "Create bucket named ``name``"
        conn = self._get_conn()
        try:
            b = conn.create_bucket(name)
            log.info("Created bucket: %s" % name)
        except BotoServerError, e:
            raise FunnelError("Bucket could not be created: %s" % name, key=name)

    def drop_bucket(self, name, **config):
        "Delete bucket named ``name``"
        conn = self._get_conn()
        try:
            conn.delete_bucket(name)
            log.info("Deleted bucket: %s" % name)
        except BotoServerError, e:
            raise FunnelError("Bucket could not be deleted: %s" % name, key=name)

    def list_bucket(self, name, marker=None, prefix=None, delimiter=None, **config):
        "Return an iterator over all the keys in bucket named ``name``"
        bucket = self._get_bucket(name)

        marker = marker or config.get('list_marker') or ''
        prefix = prefix or config.get('list_prefix') or ''
        delimiter = delimiter or config.get('list_delimiter') or ''

        more_results = True
        k = None

        log.info("Listing keys from marker: %s" % marker)
        while more_results:
            try:
                r = bucket.get_all_keys(marker=marker, prefix=prefix, delimiter=delimiter)
                for k in r:
                    yield k.name
                if k:
                    marker = k.name
                more_results= r.is_truncated
            except BotoServerError, e:
                raise FunnelError("Failed to list bucket: %s" % name, key=name)
            except (IncompleteRead, SocketError, BotoClientError), e:
                log.warning("Caught exception: %r.\nRetrying..." % e)
                time.sleep(1.0)

        log.info("Done listing bucket: %s" % name)

    def delete(self, bucket, ikeys, retry=5, **config):
        """
        Given an iterator of key names, delete these keys from the current bucket.
        Return a list of failed keys (if any).
        """
        # Setup local config for this request
        c = {}
        c.update(config)
        c['retry'] = retry

        failed = Queue()
        pool = self._get_pool()
        for k in ikeys:
            j = DeleteJob(bucket, k, failed, c)
            pool.put(j)
        pool.join()

        return collapse_queue(failed)

    def get(self, bucket, ikeys, retry=5, **config):
        """
        Given an iterator of key names, download these files from the current bucket.
        Return a list of failed keys (if any).
        """
        # Setup local config for this request
        c = {}
        c.update(config)
        c['retry'] = retry

        failed = Queue()
        pool = self._get_pool()
        for k in ikeys:
            j = GetJob(bucket, k, failed, c)
            pool.put(j)
        pool.join()

        return collapse_queue(failed)

    def put(self, bucket, ikeys, retry=5, acl='public-read', **config):
        """
        Given an iterator of file paths, put these files into the current bucket.
        Return a list of failed keys (if any).
        """
        # Setup local config for this request
        c = {}
        c.update(config)
        c['retry'] = retry
        c['acl'] = acl

        failed = Queue()
        pool = self._get_pool()
        for k in ikeys:
            j = PutJob(bucket, k, failed, c)
            pool.put(j)
        pool.join()

        return collapse_queue(failed)

    def copy(self, bucket, ikeys, retry=5, acl='public-read', **config):
        """
        Given an iterator of file paths, copy these files into the current bucket from source bucket
        Return a list of failed keys (if any).
        """
        # Setup local config for this request
        c = {}
        c.update(config)
        c['retry'] = retry
        c['acl'] = acl

        failed = Queue()
        pool = self._get_pool()
        for k in ikeys:
            j = CopyJob(bucket, k, failed, c)
            pool.put(j)
        pool.join()

        return collapse_queue(failed)



class FunnelError(Exception):
    """
    Base exception for all s3funnel errors

    Contains a ``message`` and optionally a ``key`` name for the key which
    the error is associated with. This is relative to the type of request --
    may be a bucket name or an item key inside a bucket.
    """
    def __init__(self, message, key=None):
        self._message = message
        self.key = key
    def __str__(self):
        return repr(self._message)


from setuptools import setup, find_packages
import sys, os

version = '0.6.3'

setup(name='s3funnel',
      version=version,
      description="Multithreaded tool for performing operations on Amazon's S3",
      long_description="""\
This tool uses the workerpool for multithreading and boto for access to the Amazon S3 API.""",
      classifiers=[], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='',
      author='Andrey Petrov',
      author_email='andrey.petrov@shazow.net',
      maintainer='Sorin Stoiana',
      maintainer_email='sorinalin.stoiana@gmail.com',
      url='https://github.com/sstoiana/s3funnel',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          # -*- Extra requirements: -*-
          'boto >= 2.2.2','workerpool >= 0.9.2',
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      scripts=['scripts/s3funnel'],
      )
